import sys
import util
sys.setrecursionlimit(10000)

# Modeling the problem

class TransportationProblem(object):
    def __init__(self, N: int, weights: dict[str, float]):
        self.N = N
        self.weights = weights
    def startState(self):
        return 1
    def isEnd(self, state: int) -> bool:
        return state == self.N
    def succAndCost(self, state: int) -> list[tuple[str, int, float]]:
        # return list of (action, newState, cost) triples
        result = []
        if state * 2 <= self.N:
            result.append(('tram', state * 2, self.weights['tram']))
        if state + 1 <= self.N:
            result.append(('walk', state + 1, self.weights['walk']))
        return result

# Inference

def backtrackingSearch(problem: TransportationProblem) -> tuple[float, list[tuple[str, int, float]]]:
    def recurse(state):
        if problem.isEnd(state):
            return 0, []
        minCost = float('inf')
        minHistory = []
        for action, newState, cost in problem.succAndCost(state):
            futureCost, history = recurse(newState)
            if cost+futureCost < minCost:
                minCost = cost + futureCost
                minHistory = [(action, newState, cost)] + history
        return minCost, minHistory
    return recurse(problem.startState())

def dynamicProgramming(problem: TransportationProblem) -> tuple[float, list[tuple[str, int, float]]]:
    cache = {}
    def recurse(state):
        if state not in cache:
            if problem.isEnd(state):
                cache[state] = 0, []
            else:
                minCost = float('inf')
                minHistory = []
                for action, newState, cost in problem.succAndCost(state):
                    futureCost, history = recurse(newState)
                    if cost+futureCost < minCost:
                        minCost = cost + futureCost
                        minHistory = [(action, newState, cost)] + history
                cache[state] = minCost, minHistory
        return cache[state]
    return recurse(problem.startState())

def uniformCostSearch(problem: TransportationProblem) -> tuple[float, list[tuple[str, int, float]]]:
    frontier = util.PriorityQueue()
    frontier.update(problem.startState(), 0)
    backpointer = {}
    explored = set()
    while True:
        state, priority = frontier.removeMin()
        if state == None:
            return float('inf'), []
        if problem.isEnd(state):
            history = []
            while state != problem.startState():
                prevState, action, cost = backpointer[state]
                history.append((action, state, cost))
                state = prevState
            history.reverse()
            return priority, history
        explored.add(state)
        for action, newState, cost in problem.succAndCost(state):
            if newState in explored:
                continue
            if frontier.update(newState, priority + cost):
                backpointer[newState] = (state, action, cost)

def printSolution(solution: tuple[float, list[tuple[str, int, float]]]):
    totalCost, history = solution
    print(f'totalCost: {totalCost}')
    for item in history:
        print(item, end=' ')
    print()

# Main

problem = TransportationProblem(40, {'walk': 1, 'tram': 2})
# solution = backtrackingSearch(problem)
# solution = dynamicProgramming(problem)
solution = uniformCostSearch(problem)
printSolution(solution)